#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <limits.h>

#include "m.h"
#include "cmd.h"
#include "g.h"
#include "fit.h"

static char rcsid[] = {"$Id: fitcmdfit.c,v 1.3 1997/12/15 18:54:26 kraemer Exp $"};

extern struct SFIT sFit;
/*
   --- Command processor for the FIT command

   Copyright  (C) 1987-1997 M.Kraemer, D-64283 Darmstadt, Hoelgesstr. 16,
                             all rights reserved
   $Log: fitcmdfit.c,v $
   Revision 1.3  1997/12/15 18:54:26  kraemer
   fit statistics in function structure

   Revision 1.2  1997/08/29 12:09:43  kraemer
   gfxattr rework

   Revision 1.1.1.1  1997/07/10 17:59:58  kraemer
   revision for cvs

*/

void fitCmdFit( const char *pcWdw,
                long *plLSQ, long *plPoisson, long *plZeroes,
                long *plErrAnal, long *plIter, long *plTime,
                long *plList, long *plNoDisp,
                long *plDebug
               ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__, *pMsg = pDbgMsg;
   unsigned long lFlag = FIT_NOMSG, lFlagG = 0;

   long lTime; unsigned long lBuf[2];

   struct List sLFWN, sLFDN;
   struct Node *psln;
   struct SFITDATA sFDFit, *psFD = NULL;
   struct SFITWDWNODE *psFWN = NULL;
   struct SFITWDW sFW, sFWDef, sFWZeroes, *psFW;
   int iWdw, nWdw;
   char **pvcWdw = NULL;
   int  *nvcWdw = NULL;

   struct SFITGFX sFGfx;
   struct SGFXATTR sGA;

   long lIter = 0;
   int bOK;
   double dRelChange, dChisqOld = DBL_MAX;
#define MSTR 132
   char cStr[MSTR+1];


   if ( *plDebug ) lFlag  |= FIT_DEBUG;
   if ( *plDebug ) lFlagG |= G_DEBUG;
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
   NewList( &sLFWN );
   NewList( &sLFDN );
   memset( &sFGfx, 0, sizeof(sFGfx) );
   memset( &sFDFit, 0, sizeof(sFDFit) );
                                      /* check data & function defined */
   iRC = ERRFIT_NOFUNCTION;
   if ( sFit.sFFB.nf <= 0 ) goto gError;
   iRC = ERRFIT_NODATA;
   if ( sFit.sFD.nn <= 0 ) goto gError;
                                      /* time in msec */
   if ( (lTime = *plTime * 1000) <= 0 ) lTime = LONG_MAX;
/* ??? get display scale
       (done by caller)
  - _GQueryScale( sFit.pGH, lScale, lFlagG )
  - lScale -> sFit.sFD.lScale ( GSCALE_* -> FITSCALE_* )
*/
                                      /* check window interface exists */
   if ( sFit.eFitWdw == NULL ) {
      iRC = ERRFIT_NOWDWAVAIL; goto gError;
   }
                                      /* fetch default limits */
   fitWdwInit( &sFWDef, lFlag );
   if ( sFit.eFitWdw( "#", 1, &sFWDef, lFlag ) < 0 ) {
      pMsg = "#";
      iRC = ERRFIT_NOSUCHWDW; goto gError;
   }
                                      /* fit statistics */
   sFit.sFFB.sFS.lStatus &= ~FITSTAT_POISSON;
   if ( sFit.lStatus & FIT_POISSON ) sFit.sFFB.sFS.lStatus |= FITSTAT_POISSON;
   if ( *plLSQ     ) sFit.sFFB.sFS.lStatus &= ~FITSTAT_POISSON;
   if ( *plPoisson ) {
      pMsg = "poisson"; iRC = ERRFIT_CMDPARNOTYET; goto gError;
/*
      sFit.sFFB.sFS.lStatus |=  FITSTAT_POISSON;
*/
   }
/*
   --- collect windows
*/
                                      /* window list specified */
   while ( *pcWdw == ' ' ) pcWdw++;
   if ( *pcWdw ) {
      if ( ( nWdw = c2vc( (char *) pcWdw, NULL, 1, INT_MAX ) ) <= 0 ) {
         pMsg = (char *) pcWdw;
         iRC = ERRFIT_BADWDWLISTSPEC; goto gError;
      }
      if ( ( pvcWdw = calloc( nWdw, sizeof(char **) ) ) == NULL ) {
         iRC = ERRFIT_SYS; goto gError;
      }
      if ( ( nvcWdw = calloc( nWdw, sizeof(int *) ) ) == NULL ) {
         iRC = ERRFIT_SYS; goto gError;
      }
      if ( c2vcn( (char *) pcWdw, pvcWdw, nvcWdw, 1, nWdw ) != nWdw ) {
         pMsg = (char *) pcWdw;
         iRC = ERRFIT_BADWDWLISTSPEC; goto gError;
      }
                                      /* fetch window limits from caller */
      for ( iWdw=0; iWdw<nWdw; iWdw++ ) {
         memset( &sFW, 0, sizeof(sFW) );
         if ( sFit.eFitWdw( pvcWdw[iWdw], nvcWdw[iWdw], &sFW, lFlag ) < 0 ){
            pMsg = pvcWdw[iWdw];
            iRC = ERRFIT_NOSUCHWDW; goto gError;
         }
                                      /* duplicate window node */
         if ( ( psFWN = calloc( 1, sizeof(*psFWN) ) ) == NULL ) {
            iRC = ERRFIT_SYS; goto gError;
         }
         psFWN->sFW = sFW;
                                      /* link into new list */
         AddTail( &sLFWN, &psFWN->ln ); psFWN = NULL;
      }
   } else {                           /* try fit windows */
      if ( sFit.sListFW.lh_Head -> ln_Succ ) {
                                      /* duplicate fit window list */
         if ( ( iRC = fitWdwListDup( &sFit.sListFW, &sLFWN, lFlag ) ) != 0 )
            goto gError;
         else;
      } else {                        /* at least display region */
                                      /* duplicate window */
         if ( ( psFWN = calloc( 1, sizeof(*psFWN) ) ) == NULL ) {
            iRC = ERRFIT_SYS; goto gError;
         }
         psFWN->sFW = sFWDef;
                                      /* link into new list */
         AddTail( &sLFWN, &psFWN->ln ); psFWN = NULL;
      }
   }
                                      /* ensure display region only */
   for( psln = sLFWN.lh_Head; psln->ln_Succ; psln=psln->ln_Succ ) {
      psFW = &( ( (struct SFITWDWNODE *) psln ) -> sFW );
      if ( ( iRC = fitWdwAnd( &sFWDef, psFW, psFW, lFlag ) ) != 0 )
         goto gError;
   }
/*
   --- setup data nodes for display
*/
   if ( !*plNoDisp ) {
      if ( ( iRC = fitDataWdwList( &sFit.sFD, &sLFWN, &sLFDN, lFlag ) ) != 0 )
         goto gError;
                                      /* account for display scale etc */
      for( psln = sLFDN.lh_Head; psln->ln_Succ; psln=psln->ln_Succ ) {
         psFD = &( ( (struct SFITDATANODE *) psln ) -> sFD );
         psFD->sFCalX = sFit.sFCalX;
         if ( ( iRC = fitDataScale( psFD, FIT_DIMX, lFlag ) ) != 0 )
            goto gError;
      }
   }
/*
   --- setup data node for fit
*/
                                      /* include y-zeroes or not ? */
   fitWdwInit( &sFWZeroes, lFlag );
   sFWZeroes.nDim = 2 /* sFWDef.nDim */;
   if ( !*plZeroes ) sFWZeroes.dLim[FIT_DIMY][0] = DBL_MIN;
                                      /* modify local list accordingly */
   for( psln = sLFWN.lh_Head; psln->ln_Succ; psln=psln->ln_Succ ) {
      psFW = &( ( (struct SFITWDWNODE *) psln ) -> sFW );
      psFW -> nDim = sFWZeroes.nDim;
      if ( ( iRC = fitWdwAnd( &sFWZeroes, psFW, psFW, lFlag ) ) != 0 )
         goto gError;
   }
                                      /* collect actual data for fit */
   if ( ( iRC = fitDataWdwListMerge( &sFit.sFD, &sLFWN, &sFDFit, lFlag ) ) != 0 )
      goto gError;
                                      /* apply global error modes */
   if ( sFit.sFDE.lMode > 0 ) {
      sFDFit.sFDE[FIT_DIMY] = sFit.sFDE;
   } else {
      sFit.sFDE = sFDFit.sFDE[FIT_DIMY];
   }
   if ( ( iRC = fitDataErr( &sFDFit, FIT_DIMX, lFlag ) ) != 0 )
      goto gError;
   if ( ( iRC = fitDataErr( &sFDFit, FIT_DIMY, lFlag ) ) != 0 )
      goto gError;
                                      /* apply scaling */
   sFDFit.sFCalX = sFit.sFCalX;
   if ( ( iRC = fitDataScale( &sFDFit, FIT_DIMX, lFlag | FITDATA_SCALEERR) ) != 0 )
      goto gError;
   if ( ( iRC = fitDataScale( &sFDFit, FIT_DIMY, lFlag | FITDATA_SCALEERR) ) != 0 )
      goto gError;
                                      /* check degrees of freedom */
   if ( (iRC = fitDataFctBlkStatNDF( &sFDFit, &sFit.sFFB, &sFit.sFFB.sFS, lFlag )) !=0 )
      goto gError;
   if ( sFit.sFFB.sFS.nDegFree <= 0 ) {
      iRC = ERRFIT_STATNDF; goto gError;
   }
/*
   --- prepare gfx
*/
   sFGfx.pGH = _GFndWdw( sFit.pcWdw );
   if ( !*plNoDisp  && sFGfx.pGH ) {
      
      sFGfx.lFlag = lFlagG | G_COORDLOG | G_DONTPLOT;
                                      /* clip at axes */
      _GClpRct( sFGfx.pGH, NULL, lFlagG | GCLP_INT );
                                      /* set gfx attributes */
/*
      sGA = sFit.sGA; sGA.lFct = GFXATTR_FCTXOR;
      _GGfxAttrSet( sFGfx.pGH, &sFit.sGA,
                    GGFXATTR_COLOUR | GGFXATTR_FCT |
                    GGFXATTR_LMODE | GGFXATTR_LWIDTH | lFlagG );
*/
                                /* choose complement / XOR colour mode */
      sGA = sFit.sGA;
      sGA.lFct = GFXATTR_FCTXOR; sGA.dLWidth = 0E0;
      if ( sGA.lColour == GFXATTR_FGP ) sGA.lColour = GFXATTR_BGP;
      _GGfxAttrSet( sFGfx.pGH, &sGA,
                    GGFXATTR_COLOUR | GGFXATTR_FCT |
                    GGFXATTR_LMODE | GGFXATTR_LWIDTH | lFlagG );
   }
/*
   --- here's the iteration loop
*/

   if ( *plList )
      _prtcl("<I> Iter #    Chisq/ndf log(LikeliHood)       Lambda",
             CMDPRTCL_FILE | CMDPRTCL_TERM );

   bOK = ( lIter < *plIter );
   while (bOK) {
                                      /* initiate time measurement (msec.s) */
      lBuf[1] = lBuf[0] = 0; itaskt(1000,lBuf);
                                      /* call the optimizer */
/*
      if ( sFit.sFFB.sFS.lStatus & FITSTAT_POISSON ) {
         iRC = fit1dpoi( &sFDFit, &sFit.sFFB, &sFit.sFFB.sFS, lFlag );
      } else {
*/
         iRC = fit1dlsq( &sFDFit, &sFit.sFFB, &sFit.sFFB.sFS, lFlag );
/*
      }
*/
      if ( iRC != 0 ) goto gError;
      if ( *plList ) {
         sprintf( cStr, "    %6d %12G %15G %12G",
                        lIter,
                        sFit.sFFB.sFS.dChisq / (double) sFit.sFFB.sFS.nDegFree,
                        sFit.sFFB.sFS.dLogLike,
                        sFit.sFFB.sFS.dLambda
                );
         _prtcl ( cStr, CMDPRTCL_TERM | CMDPRTCL_FILE | CMDPRTCL_NOTIME );
      }
                                      /* monitor change */
      bOK = ( ++lIter < *plIter ) && ( (dChisqOld - sFit.sFFB.sFS.dChisq) >= sFit.sFFB.sFS.dEps[2] * dChisqOld );
      dChisqOld = sFit.sFFB.sFS.dChisq;
                                      /* display fit curve */
      if ( !*plNoDisp ) {
                                      /* loop local window/data nodes */
         for( psln = sLFDN.lh_Head; psln->ln_Succ; psln=psln->ln_Succ ) {
                                      /* animation only on screen */
            psFD = &( ( (struct SFITDATANODE *) psln ) -> sFD );
            if ( _GWdwIsScreen(sFGfx.pGH) > 0 ) {
                                      /* first erase old curve */
               if ( lIter > 1 ) {
                  if ( (iRC = fitDataGfxLine( psFD, &sFGfx, lFlag )) != 0 )
                     goto gError;
               }
            }
                                      /* final curve for all */
            if ( (_GWdwIsScreen(sFGfx.pGH) > 0) || !bOK ) {
                                      /* compute new */
               if ( (iRC = fitDataFctBlkValue( psFD, &sFit.sFFB, lFlag )) != 0)
                  goto gError;
                                      /* leaving iteration: no complement */
               if (!bOK) {
                 sGA = sFit.sGA;
                 _GGfxAttrSet( sFGfx.pGH, &sGA,
                               GGFXATTR_FCT | GGFXATTR_COLOUR |
                               lFlagG );
               }
                                      /* display new */
               if ( (iRC = fitDataGfxLine( psFD, &sFGfx, lFlag )) != 0 )
                  goto gError;
            }
         }
      }
                                      /* time measurement (msec.s) */
      if ( (lTime -= itaskt(1000,lBuf)) <= 0 ) {
         bOK = 0;
         sprintf( cStr, "<W> fitdo: timeout (%d sec) expired", *plTime );
         _prtcl( cStr, CMDPRTCL_TERM | CMDPRTCL_FILE );
      }
   }
                                      /* delete old, store new data */
   fitDataFree( &sFit.sFDFit, lFlag );
   sFit.sFDFit = sFDFit; memset( &sFDFit, 0, sizeof(sFDFit) );
                                      /* remember lsq/poisson fit */
   sFit.lStatus &= ~FIT_POISSON;
   if ( sFit.sFFB.sFS.lStatus & FITSTAT_POISSON ) sFit.lStatus |= FIT_POISSON;

   iRC = 0;
gError:;
   if ( (iRC < 0) /*&& !(lFlag&FIT_NOMSG)*/ ) {
      fitemsg( pMsg, iRC );
   }
   if ( *plNoDisp == 0 ) _GClpRct( sFGfx.pGH, NULL, lFlagG | GCLP_EXT );
                                      /* cleanup */
   fitDataFree( &sFDFit, lFlag );
   fitWdwListFree( &sLFWN, lFlag );
   fitDataNodeListFree( &sLFDN, lFlag );
   if ( psFWN  ) free( psFWN  ); psFWN = NULL;
   if ( pvcWdw ) free( pvcWdw ); pvcWdw = NULL;
   if ( nvcWdw ) free( nvcWdw ); nvcWdw = NULL;

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return;
}

