
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <limits.h>

#include "u.h"
#include "m.h"
#include "g.h"
#include "fit.h"

static char rcsid[] = {"$Id: fitcmdlike.c,v 1.1.1.1 1997/07/10 17:59:59 kraemer Exp $"};

extern struct SFIT sFit;
/*
   --- Command processor for the FLIKE command

   Copyright  (C) 1987-1997 M.Kraemer, D-64283 Darmstadt, Hoelgesstr. 16,
                             all rights reserved
   $Log: fitcmdlike.c,v $
   Revision 1.1.1.1  1997/07/10 17:59:59  kraemer
   revision for cvs

*/

void fitCmdLike(
                const char *pcAnl, const char *pcEvt,
                const char *pcPar, const char *pcLim,
                const char *pcSteps,
                long *plLog,
                long *plDebug
               ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__, *pMsg = pDbgMsg;
   unsigned long lFlag = 0;

   struct SFITDATAXFER sFDXF;
   struct SFITANLXFER  sFAXF;
   struct SFITPAR sFP[MFITLIKEPAR], sFPSave[MFITLIKEPAR];
   struct SFITSTAT sFS;
   struct SFITFCTBLK sFFB;
   struct SFITDATA *psFD;

   int iDim, nDim;
   long ll;

   int ip, np=0;
#define MVPAR (MFITLIKEPAR+1)
   long lvPar[MVPAR];
   double dvLim[MVPAR][2];
   long lvScale[MVPAR];
   int  nv[MFITLIKEPAR];
   long lv[MFITLIKEPAR];
   double dHugeVal;


   if ( *plDebug ) lFlag |= FIT_DEBUG;
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
   memset( &sFDXF,   0, sizeof(sFDXF  ) );
   memset( &sFAXF,   0, sizeof(sFAXF  ) );
   memset( &sFFB,    0, sizeof(sFFB   ) );
   memset(  sFP,     0, sizeof(sFP    ) );
   memset(  sFPSave, 0, sizeof(sFPSave) );
   memset( &sFS,     0, sizeof(sFS    ) );
   memset( lvPar,    0, sizeof(lvPar  ) );
                                      /* check prerequisites */
   if ( sFit.sFFB.nf <= 0 ) {
      iRC = ERRFIT_NOFUNCTION; goto gError;
   }
                                      /* */
   psFD = &sFit.sFDFit;
   if ( psFD->nn <= 0 ) psFD = &sFit.sFD;
   if ( psFD->nn <= 0 ) {
      iRC = ERRFIT_NODATA; goto gError;
   }
                                      /* check store interface exists */
   if ( sFit.eFitLike == NULL ) {
      iRC = ERRFIT_STORENOTAVAIL; goto gError;
   }
/*
   --- get parameter number(s)
*/
   iRC = c2vl( pcPar, lvPar, 1, MVPAR );
   switch( iRC ) {
   case 0:                            /* first non-fixed parameter */
      for ( ip=0; ip<sFit.sFFB.np; ip++ ) {
         if ( iRC = fitFctBlkParGetByNum( &sFit.sFFB, ip+1, sFP+0, lFlag ) )
            goto gError;
         if ( !(sFP[0].lFlag & FITPAR_FIX) ) {
            np = 1; lvPar[0] = ip+1; break;
         }
      }
   break;
   default:
      if ( iRC < 0 ) {
         pMsg = (char *) pcPar; iRC = ERRFIT_CONVNAN; goto gError;
      }
      if ( iRC > MFITLIKEPAR ) {
         pMsg = (char *) pcPar; iRC = ERRFIT_LIKEPARTOOMANY; goto gError;
      }
                                      /* get parameter(s) */
      np = iRC;
      for ( ip=0; ip<np; ip++ ) {
         if ( iRC = fitFctBlkParGetByNum( &sFit.sFFB, lvPar[ip], sFP+ip, lFlag ) )
            goto gError;
      }
   break;
   }
   if ( np <= 0 ) {
      iRC = ERRFIT_BADPARCOUNT; goto gError;
   }
/*
   --- # of steps 
*/
                                      /* defaults */
   for ( ip=0; ip<np; ip++ ) {
      lv[ip] = MFITLIKEDIM;
   }
   iRC = c2vl( pcSteps, lv, 1, np );
   if ( iRC < 0 ) {
      pMsg = (char *) pcSteps; iRC = ERRFIT_CONVNAN; goto gError;
   }
   for ( ip=0; ip<np; ip++ ) {
      if ( ( nv[ip] = lv[ip] ) <= 0 ) {
         iRC = ERRFIT_LIKENSTEPS; goto gError;
      }
   }
/*
   --- get parameter limits and array dimension(s)
*/
                                      /* defaults */
   dHugeVal = powi( (double) FLT_RADIX, DBL_MAX_EXP-1 );
   for ( ip=0; ip<np; ip++ ) {
      if ( (dvLim[ip][0] = sFP[ip].dLim[0]) == -DBL_MAX )
         dvLim[ip][0] = -dHugeVal;
      if ( (dvLim[ip][1] = sFP[ip].dLim[1]) ==  DBL_MAX )
         dvLim[ip][1] =  dHugeVal;
   }
                                      /* callback to query limits/dim. */

                                      /* use specified limits */
   iRC = c2vd( pcLim, (double *) dvLim, 1, sizeof(dvLim)/sizeof(double) );
   switch(iRC) {
   case 0: break;                     /* none */
   default:
      if ( iRC < 0 ) {
         pMsg = (char *) pcLim; iRC = ERRFIT_CONVNAN; goto gError;
      }
      if ( iRC != (np*2) ) {
         pMsg = (char *) pcLim; iRC = ERRFIT_LIKELIMINCONSISTENT; goto gError;
      }
   break;
   }
                                      /* allocate workspace */
   if ( iRC = fitDataXFerAlloc( nv, np, &sFDXF, lFlag ) ) goto gError;
                                      /* fill base arrays */
   for ( ip=0; ip<np; ip++ ) {
      lvScale[ip] = FITSCALE_LIN;
      if ( *plLog ) {
         if ( ( dvLim[ip][0] > 0E0 ) &&
              ( dvLim[ip][1] > 0E0 ) ) {
            dvLim[ip][0] = log10( dvLim[ip][0] );
            dvLim[ip][1] = log10( dvLim[ip][1] );
            lvScale[ip] = FITSCALE_LOG;
         }
      }
      vrampd( min( dvLim[ip][0], dvLim[ip][1] ),
              max( dvLim[ip][0], dvLim[ip][1] ),
              sFDXF.pd[ip][FITXFER_DATA], 1, sFDXF.pl[ip] );
      if ( lvScale[ip] == FITSCALE_LOG )
         v10tox( sFDXF.pd[ip][FITXFER_DATA], 1,
                 sFDXF.pd[ip][FITXFER_DATA], 1,
                 sFDXF.pl[ip] );
   }
/*
   --- compute likelihood function
*/
   memset( lv, 0, sizeof(lv) );       /* clear index counter */
                                      /* duplicate function block */
   if ( iRC = fitFctBlkDup( &sFit.sFFB, &sFFB, lFlag ) ) goto gError;
                                      /* loop over data dimension */
   nDim = sFDXF.nDim;
   for ( ll=0; ll<sFDXF.pl[ nDim ]; ll++ ) {
                                      /* update parameter values */
      for ( ip=0; ip<np; ip++ ) {
         sFP[ip].dVal = sFDXF.pd[ip][FITXFER_DATA][ lv[ip] ];
         if ( iRC = fitFctBlkParPutByNum( &sFFB, lvPar[ip], sFP+ip, lFlag ) )
            goto gError;
      }
                                      /* compute function values */
      if ( iRC = fitDataFctBlkValue( psFD, &sFFB, lFlag ) ) goto gError;
                                      /* compute likelihood */
      if ( iRC = fitDataStatChi( psFD, &sFS, lFlag ) ) goto gError;
      sFDXF.pd[nDim][FITXFER_DATA][ll] = sFS.dChisq; /*exp( sFS.dLogLike);*/
                                      /* */
      ++lv[ nDim-1 ];                 /* increment fastest index */
      for ( iDim=nDim-1; iDim>=0; iDim-- ) {
                                      /* increment slower indices */
         if ( lv[iDim] >= sFDXF.pl[ iDim ] ) {
            lv[iDim] = 0; if ( iDim > 0 ) lv[iDim-1] ++;
         }
      }
   }
/*
   --- callback to store results
*/
   while( *pcAnl == ' ' ) pcAnl++;
   if ( ( sFAXF.pcAnl = strdup(pcAnl) ) == NULL ) {
      iRC = ERRFIT_SYS; pMsg = (char *) pcAnl; goto gError;
   }
   while( *pcEvt == ' ' ) pcEvt++;
   if ( ( sFAXF.pcEvt = strdup(pcEvt) ) == NULL ) {
      iRC = ERRFIT_SYS; pMsg = (char *) pcEvt; goto gError;
   }
                                      /* */
   if ( sFit.eFitLike( &sFAXF, &sFDXF, lFlag | FITCBACK_LIKE ) < 0 ) {
      iRC = ERRFIT_STOREFAILED; goto gError;
   }
                                      /*  */
   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }

   fitFctBlkFree( &sFFB, lFlag );
   fitAnlXFerFree( &sFAXF, lFlag );
   fitDataXFerFree( &sFDXF, lFlag );

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return;
}
