#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <limits.h>

#include "u.h"
#include "m.h"
#include "g.h"
#include "fit.h"

static char rcsid[] = {"$Id: fitcmdstore.c,v 1.1.1.1 1997/07/10 18:00:01 kraemer Exp $"};

extern struct SFIT sFit;
/*
   --- Command processor for the FSTORE command

   Copyright  (C) 1987-1997 M.Kraemer, D-64283 Darmstadt, Hoelgesstr. 16,
                             all rights reserved
   $Log: fitcmdstore.c,v $
   Revision 1.1.1.1  1997/07/10 18:00:01  kraemer
   revision for cvs

*/

void fitCmdStore(
                 const char *pcAnl, const char *pcEvt,
                 double *pdConf,
                 long *plDiff, long *plChi, long *plNoConfirm,
                 long *plDebug
                ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__, *pMsg = pDbgMsg;
   unsigned long lFlag = 0;
   int ii;

   int nv[] = {0}, nDim;
   struct SFITDATAXFER sFDXF;
   struct SFITANLXFER  sFAXF;


   if ( *plDebug ) lFlag |= FIT_DEBUG;
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
   memset( &sFDXF, 0, sizeof(sFDXF) );
   memset( &sFAXF, 0, sizeof(sFAXF) );
                                      /* check valid inputs */
   if ( *pdConf >= 1E0 ) {
      iRC = ERRFIT_INVCONF; goto gError;
   }
   if ( *pdConf <= 0E0 ) {
      iRC = ERRFIT_INVCONF; goto gError;
   }
                                      /* check prerequisites */
   if ( sFit.sFFB.nf <= 0 ) {
      iRC = ERRFIT_NOFUNCTION; goto gError;
   }
   if ( sFit.sFDFit.nn <= 0 ) {
      iRC = ERRFIT_NODATA; goto gError;
   }
                                      /* check store interface exists */
   if ( sFit.eFitStore == NULL ) {
      iRC = ERRFIT_STORENOTAVAIL; goto gError;
   }
                                      /* prepare transfer structure(s) */
   nv[0] = sFit.sFDFit.nn;
   if ( iRC = fitDataXFerAlloc( nv, nDim=1, &sFDXF, lFlag ) ) goto gError;
                                      /* copy x-arrays */
   memcpy( sFDXF.pd[0][FITXFER_DATA],
           sFit.sFDFit.pdD [FIT_DIMX][FITDATA_DATA], sFDXF.pl[0] * sizeof(double) );
   memcpy( sFDXF.pd[0][FITXFER_ERRM],
           sFit.sFDFit.pdD [FIT_DIMX][FITDATA_ERRM], sFDXF.pl[0] * sizeof(double) );
   memcpy( sFDXF.pd[0][FITXFER_ERRP],
           sFit.sFDFit.pdD [FIT_DIMX][FITDATA_ERRP], sFDXF.pl[0] * sizeof(double) );
                                      /* compute function+errors */
   if ( iRC = fitDataFctBlkConf( &sFit.sFDFit, &sFit.sFFB, *pdConf, lFlag ) ) goto gError;
                                      /* scale for display scale */
   if ( iRC = fitDataScaleFct( &sFit.sFDFit, FITDATA_SCALEFCT, lFlag | FITDATA_SCALEERR ) ) goto gError;
                                      /* copy yfct-arrays */
   memcpy( sFDXF.pd[1][FITXFER_DATA],
           sFit.sFDFit.pdFS[FITDATA_DATA], sFDXF.pl[1] * sizeof(double) );
   memcpy( sFDXF.pd[1][FITXFER_ERRM],
           sFit.sFDFit.pdFS[FITDATA_ERRM], sFDXF.pl[1] * sizeof(double) );
   memcpy( sFDXF.pd[1][FITXFER_ERRP],
           sFit.sFDFit.pdFS[FITDATA_ERRP], sFDXF.pl[1] * sizeof(double) );
                                      /* (y-yfct) */
   if ( *plDiff ) {
      if ( iRC = fitDataNetTrue( &sFit.sFDFit, lFlag ) ) goto gError;
      memcpy( sFDXF.pd[1][FITXFER_DATA],
              sFit.sFDFit.pdRR[FIT_DIMY][FITDATA_DATA], sFDXF.pl[1] * sizeof(double) );
      memcpy( sFDXF.pd[1][FITXFER_ERRM],
              sFit.sFDFit.pdRR[FIT_DIMY][FITDATA_ERRM], sFDXF.pl[1] * sizeof(double) );
      memcpy( sFDXF.pd[1][FITXFER_ERRP],
              sFit.sFDFit.pdRR[FIT_DIMY][FITDATA_ERRP], sFDXF.pl[1] * sizeof(double) );
   }
                                      /* (y-yfct)/error */
   if ( *plChi ) {
      if ( iRC = fitDataNetTrueNorm( &sFit.sFDFit, lFlag ) ) goto gError;
      memcpy( sFDXF.pd[1][FITXFER_DATA],
              sFit.sFDFit.pdRR[FIT_DIMY][FITDATA_DATA], sFDXF.pl[1] * sizeof(double) );
      memcpy( sFDXF.pd[1][FITXFER_ERRM],
              sFit.sFDFit.pdRR[FIT_DIMY][FITDATA_ERRM], sFDXF.pl[1] * sizeof(double) );
      memcpy( sFDXF.pd[1][FITXFER_ERRP],
              sFit.sFDFit.pdRR[FIT_DIMY][FITDATA_ERRP], sFDXF.pl[1] * sizeof(double) );
   }
                                      /* copy analyzer/event name */
   while( *pcAnl == ' ' ) pcAnl++;
   if ( ( sFAXF.pcAnl = strdup(pcAnl) ) == NULL ) {
      iRC = ERRFIT_SYS; pMsg = (char *) pcAnl; goto gError;
   }
   while( *pcEvt == ' ' ) pcEvt++;
   if ( ( sFAXF.pcEvt = strdup(pcEvt) ) == NULL ) {
      iRC = ERRFIT_SYS; pMsg = (char *) pcEvt; goto gError;
   }
                                      /* */
   if ( sFit.eFitStore( &sFAXF, &sFDXF, lFlag ) < 0 ) {
      iRC = ERRFIT_STOREFAILED; goto gError;
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
                                      /* cleanup */
   fitDataXFerFree( &sFDXF, lFlag );
   fitAnlXFerFree ( &sFAXF, lFlag );

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return;
}
