#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <limits.h>

#include "u.h"
#include "m.h"
#include "polfit.h"
#include "cmd.h"
#include "g.h"
#include "fit.h"

static char rcsid[] = {"$Id: fitcmdxcal.c,v 1.1.1.1 1997/07/10 18:00:02 kraemer Exp $"};

extern struct SFIT sFit;
/*
   --- Command processor for the FXCAL command

   Copyright  (C) 1987-1997 M.Kraemer, D-64283 Darmstadt, Hoelgesstr. 16,
                             all rights reserved
   $Log: fitcmdxcal.c,v $
   Revision 1.1.1.1  1997/07/10 18:00:02  kraemer
   revision for cvs

*/

void fitCmdXCal(
                double *pdOffset, double *pdFactor,
                const char *pcChannels, const char *pcXNew,
                const char *pcErrors, const char *pcAnl,
                long *plDebug
               ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__, *pMsg = pDbgMsg;
   unsigned long lFlag = 0;

   unsigned long lFlagPF = 0;
   struct SPOLFITDATA sPFD;
   struct SPOLFITPAR  sPFP;

#define MPAR 2
   double dvPar[MPAR];

   int ii, nn; double dd;
   double *pvdCh = NULL, *pvdXN = NULL /*, *pvdW = NULL*/;

   char cStr[255+1];


   if ( *plDebug ) lFlag |= FIT_DEBUG;
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
   memset( &sPFD, 0, sizeof(sPFD) );
   memset( &sPFP, 0, sizeof(sPFP) );
   if ( lFlag & FIT_DEBUG ) lFlagPF |= POLFIT_DEBUG;

   dvPar[0] = *pdOffset;
   if ( (dvPar[1] = *pdFactor) <= 0E0 ) {
      iRC = ERRFIT_XCALSLOPE; goto gError;
   }
                                      /* strip leading blanks */
   while( *pcChannels == ' ' ) pcChannels++;
   while( *pcXNew     == ' ' ) pcXNew++;
   while( *pcErrors   == ' ' ) pcErrors++;
                                      /* some channels given ? */
   if ( *pcChannels ) {
      nn = iRC = c2vd( pcChannels, pvdCh, 1, INT_MAX );
      if ( iRC < 0 ) {
         iRC = ERRFIT_CONVNAN; pMsg = (char *) pcChannels; goto gError;
      }
      if ( iRC < 2 ) {
         iRC = ERRFIT_XCALFEWCHN; pMsg = (char *) pcChannels; goto gError;
      }
                                      /* new x-values to map */
      iRC = c2vd( pcXNew, pvdXN, 1, INT_MAX );
      if ( iRC < 0 ) {
         iRC = ERRFIT_CONVNAN; pMsg = (char *) pcXNew; goto gError;
      }
      if ( iRC < 2 ) {
         iRC = ERRFIT_XCALFEWVAL; pMsg = (char *) pcXNew; goto gError;
      }
                                      /* check consistency */
      if ( iRC < nn ) {
         iRC = ERRFIT_XCALFEWVAL; pMsg = (char *) pcXNew; goto gError;
      }
      if ( iRC > nn ) {
         iRC = ERRFIT_XCALFEWCHN; pMsg = (char *) pcChannels; goto gError;
      }
                                      /* space for channels, values */
      if ( (pvdCh = calloc( nn, sizeof(double) )) == NULL ) {
         iRC = ERRFIT_SYS; goto gError;
      }
      c2vd( pcChannels, pvdCh, 1, nn );
      if ( (pvdXN = calloc( nn, sizeof(double) )) == NULL ) {
         iRC = ERRFIT_SYS; goto gError;
      }
      c2vd( pcXNew, pvdXN, 1, nn );
                                      /* compute coeff. directly */
      if ( nn == 2 ) {
                                      /* check increasing channels */
         if ( ( pvdCh[1] - pvdCh[0] ) == 0E0 ) {
            pMsg = (char *) pcChannels; iRC = ERRFIT_XCALCHN; goto gError;
         }
                                      /* need to swap ? */
         if ( ( pvdCh[1] - pvdCh[0] ) < 0E0 ) {
            dd = pvdCh[1]; pvdCh[1] = pvdCh[0]; pvdCh[0] = dd;
            dd = pvdXN[1]; pvdXN[1] = pvdXN[0]; pvdXN[0] = dd;
         }
         dvPar[1] = ( pvdXN[1] - pvdXN[0] ) / ( pvdCh[1] - pvdCh[0] );
         if ( dvPar[1] <= 0E0 ) {
            iRC = ERRFIT_XCALSLOPE; goto gError;
         }
         dvPar[0] = pvdXN[0] - pvdCh[0] * dvPar[1];

      } else {                        /* least squares fit */
                                      /* space for weights etc. */
         iRC = polfitParAlloc( MPAR, &sPFP, lFlagPF );
         if ( iRC != 0 ) {
            iRC = ERRFIT_XCALFITPOL; goto gError;
         }
         iRC = polfitDataAlloc( nn, &sPFD, lFlagPF );
         if ( iRC != 0 ) {
            iRC = ERRFIT_XCALFITPOL; goto gError;
         }
         vmov( pvdCh, 1, sPFD.pdX, 1, nn );
         vmov( pvdXN, 1, sPFD.pdY, 1, nn );
                                      /* errors given ? */
         if ( *pcErrors ) {
            if ( (iRC = c2vd( pcErrors, NULL, 1, INT_MAX )) < 0 ) {
               pMsg = (char *) pcErrors; iRC = ERRFIT_CONVNAN; goto gError;
            }
                                      /* check consistency */
            if ( iRC < nn ) {
               iRC = ERRFIT_XCALFEWERR; pMsg = (char *) pcXNew; goto gError;
            }
            c2vd( pcErrors, sPFD.pddYm, 1, nn );
            c2vd( pcErrors, sPFD.pddYp, 1, nn );
                                      /* set weights */
         }

         iRC = polfitFit( &sPFD, &sPFP, lFlagPF );
         if ( iRC != 0 ) {
            iRC = ERRFIT_XCALFITPOL; goto gError;
         }
         dvPar[0] = sPFP.pdPar[0];
         dvPar[1] = sPFP.pdPar[1];
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> chisq=%G\n", sPFP.dChisq );
      }
   }

   sFit.sFCalX.dCoeff[0] = dvPar[0];
   sFit.sFCalX.dCoeff[1] = dvPar[1];

   sprintf( cStr, "<I> xvalues will be converted according to %G + %G * xvalue",
                  sFit.sFCalX.dCoeff[0], sFit.sFCalX.dCoeff[1] );
   _prtcl( cStr, CMDPRTCL_TERM | CMDPRTCL_FILE );

if ( lFlag & FIT_DEBUG ) fitCalInspect( &sFit.sFCalX, lFlag );
   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }

   polfitParFree( &sPFP, lFlagPF );
   polfitDataFree( &sPFD, lFlagPF );
   if ( pvdCh ) free( pvdCh ); pvdCh = NULL;
   if ( pvdXN ) free( pvdXN ); pvdXN = NULL;
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return;
}
