#ifndef MATHWDW_H
#define MATHWDW_H
/*
   
   Definitions for coordinate windows

   Copyright  (C) 1992-2000 M.Kraemer, D-64283 Darmstadt, Hoelgesstr. 16,
                             all rights reserved

   $Log$

*/

/*

  General
  =======

  o organization of multi-dim polygons 
      
*/

#include "u.h"
#include "m.h"
                                      /* masks */
#define MATHWDW_MASKGLOBAL 0xFF000000 /* global (e.g. debug) */
#define MATHWDW_MASKFUNC   0x00FF0000 /* function specific flags */
#define MATHWDW_MASKQUAL   0x0000FFFF /* for additional qualifiers */
                                      /* flags */
#define MATHWDW_DEBUG      0x80000000 /* general check flag */
#define MATHWDW_CHECK      MATHWDW_DEBUG  /* general check flag */
#define MATHWDW_QUIET      0x10000000 /* suppress error messages */
#define MATHWDW_NOMSG      MATHWDW_QUIET
                                      /* */
#define MATHWDWDIM 4
#define MATHWDWMOM 4
#define MATHWDWINC 16

struct SMATHWDW {
   double dvLim[MATHWDWDIM][2];       /* rectangular limits */
                                      /* sum, mean, variance, skewness */
   double dvMom[MATHWDWDIM][MATHWDWMOM];
   unsigned long ulFlag; 
   unsigned long ulType; 
   int nDim;                          /* actual dimensionality */
   int iPntDim[2];                    /* dimensions to which pnts belong */
   int mPnt;                          /* max # of 2D polygon points */ 
   int nPnt;                          /* current # of 2D polygon points */ 
   double **pdPnt;                    /* [nPnt][2] 2D polygon points */
};
                                      /* */
enum {
   MATHWDW_DIMX=0, MATHWDW_DIMY, MATHWDW_DIMZ, 
};                                            
enum {
   MATHWDW_BELOW=1, MATHWDW_ABOVE=2,  
};                                            

#define MATHWDW_REFCOPY 0x000000001   /* do not free() */

                                      /* macro for rectangular wdw check */
#define MATHWDWRECTINSIDEEXCL(iRC,psW,dv,mDim) \
{ int iMATHWDWDim, nMATHWDWDim;\
nMATHWDWDim = min( mDim, MATHWDWDIM );\
iRC = 1;\
for ( iMATHWDWDim=0; iMATHWDWDim<nMATHWDWDim; iMATHWDWDim++ ) {\
  if ( dv[iMATHWDWDim] <= (psW) -> dvLim[iMATHWDWDim][0] ) { iRC = 0; break; }\
  if ( dv[iMATHWDWDim] >= (psW) -> dvLim[iMATHWDWDim][1] ) { iRC = 0; break; }\
}\
}

#define MATHWDWRECTINSIDEINCL(iRC,psW,dv,mDim) \
{ int iMATHWDWDim, nMATHWDWDim;\
nMATHWDWDim = min( mDim, MATHWDWDIM );\
iRC = 1;\
for ( iMATHWDWDim=0; iMATHWDWDim<nMATHWDWDim; iMATHWDWDim++ ) {\
  if ( dv[iMATHWDWDim] <  (psW) -> dvLim[iMATHWDWDim][0] ) { iRC = 0; break; }\
  if ( dv[iMATHWDWDim] >  (psW) -> dvLim[iMATHWDWDim][1] ) { iRC = 0; break; }\
}\
}

#define MATHWDWRECTWHEREINCL(iRC,ulvWhere,psW,dv,mDim) \
{ int iMATHWDWDim, nMATHWDWDim;\
nMATHWDWDim = min( mDim, MATHWDWDIM );\
iRC = 1;\
for ( iMATHWDWDim=0; iMATHWDWDim<nMATHWDWDim; iMATHWDWDim++ ) {\
  ulvWhere[iMATHWDWDim] = 0;\
  if ( dv[iMATHWDWDim] <  (psW) -> dvLim[iMATHWDWDim][0] ) { iRC = 0; ulvWhere[iMATHWDWDim]|=MATHWDW_BELOW; }\
  if ( dv[iMATHWDWDim] >  (psW) -> dvLim[iMATHWDWDim][1] ) { iRC = 0; ulvWhere[iMATHWDWDim]|=MATHWDW_ABOVE; }\
}\
}

/*

   Functions for window handling
   =============================

   mathWdwAndExcl: ( psWdwTo, psWdw1, psWdw2, flag )
                   o logical and of windows, exclude boundaries
                   o returns:
                     <  0: error
                     == 0: closed (no overlap)
                     >  0: opened (overlap)
                   o to test
                   ! polygons too 
                   o calls

   mathWdwAndIncl: ( psWdwTo, psWdw1, psWdw2, flag )
                   o logical and of windows, include boundaries
                   o returns:
                     <  0: error
                     == 0: closed (no overlap)
                     >  0: opened (overlap)
                   o to test
                   ! polygons too 
                   o calls

   mathWdwCopy   : ( psWTo, psWFr, flag )
                   o todo
                   o duplicate window
                   o calls
   
   mathWdwDup    : ( psWTo, psWFr, flag )
                 4 o OK
                   o duplicate window
                   o calls
   
   mathWdwFree   : ( psWdw, flag )
                 3 o OK
                   o calls
   
   mathWdwInit   : ( psWdw, flag )
                 2 o OK
                   o "open" by default
                   o calls
   
   mathWdwInsideExcl: ( psWdw, dv, dim, flag )
                      o check inside, excluding boundaries
                      o returns
                        <0  for error
                        ==0 for outside
                        >0  for inside
                      o to complete
                      ! polygon check
                      o calls
   
   mathWdwInsideIncl: ( psWdw, dv, dim, flag )
                      o check inside, including boundaries  
                      o returns
                        <0  for error
                        ==0 for outside
                        >0  for inside
                      o to complete
                      ! polygon check
                      o calls
   
   mathWdwInspect: ( psWdw, flag )
                 1 o OK
                   o calls
   
   mathWdwOr     : ( psWdwTo, psWdw1, psWdw2, flag )
                 6 o OK
                   o logical or of windows
                   o more extensive tests
                   ! polygons too 
                   o calls

   mathWdwPntAdd : ( psWdw, dv[], flag )
                 5 o OK
                   o add a point to a polygon wdw
                   o evtl: recalculate rectangular limits
                   o calls
   
   mathWdwRectLineSect: ( psWdw, pnto, dir, pnti, dim, flag )
                        o intersection w/ rectangular limits  
                        o returns
                        x to test (more tests to do ?)
                        o calls
   

*/
                                      /* function prototypes */
int mathWdwAndExcl   ( struct SMATHWDW *, struct SMATHWDW *, struct SMATHWDW *, unsigned long );
int mathWdwAndIncl   ( struct SMATHWDW *, struct SMATHWDW *, struct SMATHWDW *, unsigned long );
int mathWdwCopy      ( struct SMATHWDW *, struct SMATHWDW *, unsigned long );
int mathWdwDup       ( struct SMATHWDW *, struct SMATHWDW *, unsigned long );
int mathWdwFree      ( struct SMATHWDW *, unsigned long );
int mathWdwInit      ( struct SMATHWDW *, unsigned long );
int mathWdwInsideExcl( struct SMATHWDW *, double[], int, unsigned long );
int mathWdwInsideIncl( struct SMATHWDW *, double[], int, unsigned long );
int mathWdwInspect   ( struct SMATHWDW *, unsigned long );
int mathWdwOr        ( struct SMATHWDW *, struct SMATHWDW *, struct SMATHWDW *, unsigned long );
int mathWdwPntAdd    ( struct SMATHWDW *, double [], unsigned long );

int mathWdwRectLineSect( struct SMATHWDW *, double[], double[], double[], int, unsigned long );

struct SMATHWDW *_mathWdwTest( unsigned long );

enum { ERRMATHWDW_SYS=1,
       ERRMATHWDW_BADADDR,
       ERRMATHWDW_BADNDIM,
};                                      

#endif
